package org.apache.bcel.classfile;

import java.util.HashMap;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;
import org.apache.bcel.Const;
import java.io.DataInput;
import java.util.Map;

public abstract class Attribute implements Cloneable, Node
{
    private static final boolean debug;
    private static final Map<String, Object> READERS;
    public static final Attribute[] EMPTY_ARRAY;
    @Deprecated
    protected int name_index;
    @Deprecated
    protected int length;
    @Deprecated
    protected byte tag;
    @Deprecated
    protected ConstantPool constant_pool;
    
    @Deprecated
    public static void addAttributeReader(final String name, final AttributeReader attributeReader) {
        Attribute.READERS.put(name, attributeReader);
    }
    
    public static void addAttributeReader(final String name, final UnknownAttributeReader unknownAttributeReader) {
        Attribute.READERS.put(name, unknownAttributeReader);
    }
    
    protected static void println(final String msg) {
        if (Attribute.debug) {
            System.err.println(msg);
        }
    }
    
    public static Attribute readAttribute(final DataInput file, final ConstantPool constantPool) throws IOException {
        byte tag = -1;
        final int nameIndex = file.readUnsignedShort();
        final String name = constantPool.getConstantUtf8(nameIndex).getBytes();
        final int length = file.readInt();
        for (byte i = 0; i < 27; ++i) {
            if (name.equals(Const.getAttributeName(i))) {
                tag = i;
                break;
            }
        }
        switch (tag) {
            case -1: {
                final Object r = Attribute.READERS.get(name);
                if (r instanceof UnknownAttributeReader) {
                    return ((UnknownAttributeReader)r).createAttribute(nameIndex, length, file, constantPool);
                }
                return new Unknown(nameIndex, length, file, constantPool);
            }
            case 1: {
                return new ConstantValue(nameIndex, length, file, constantPool);
            }
            case 0: {
                return new SourceFile(nameIndex, length, file, constantPool);
            }
            case 2: {
                return new Code(nameIndex, length, file, constantPool);
            }
            case 3: {
                return new ExceptionTable(nameIndex, length, file, constantPool);
            }
            case 4: {
                return new LineNumberTable(nameIndex, length, file, constantPool);
            }
            case 5: {
                return new LocalVariableTable(nameIndex, length, file, constantPool);
            }
            case 6: {
                return new InnerClasses(nameIndex, length, file, constantPool);
            }
            case 7: {
                return new Synthetic(nameIndex, length, file, constantPool);
            }
            case 8: {
                return new org.apache.bcel.classfile.Deprecated(nameIndex, length, file, constantPool);
            }
            case 9: {
                return new PMGClass(nameIndex, length, file, constantPool);
            }
            case 10: {
                return new Signature(nameIndex, length, file, constantPool);
            }
            case 11: {
                println("Warning: Obsolete StackMap attribute ignored.");
                return new Unknown(nameIndex, length, file, constantPool);
            }
            case 12: {
                return new RuntimeVisibleAnnotations(nameIndex, length, file, constantPool);
            }
            case 13: {
                return new RuntimeInvisibleAnnotations(nameIndex, length, file, constantPool);
            }
            case 14: {
                return new RuntimeVisibleParameterAnnotations(nameIndex, length, file, constantPool);
            }
            case 15: {
                return new RuntimeInvisibleParameterAnnotations(nameIndex, length, file, constantPool);
            }
            case 16: {
                return new AnnotationDefault(nameIndex, length, file, constantPool);
            }
            case 17: {
                return new LocalVariableTypeTable(nameIndex, length, file, constantPool);
            }
            case 18: {
                return new EnclosingMethod(nameIndex, length, file, constantPool);
            }
            case 19: {
                return new StackMap(nameIndex, length, file, constantPool);
            }
            case 20: {
                return new BootstrapMethods(nameIndex, length, file, constantPool);
            }
            case 21: {
                return new MethodParameters(nameIndex, length, file, constantPool);
            }
            case 22: {
                return new Module(nameIndex, length, file, constantPool);
            }
            case 23: {
                return new ModulePackages(nameIndex, length, file, constantPool);
            }
            case 24: {
                return new ModuleMainClass(nameIndex, length, file, constantPool);
            }
            case 25: {
                return new NestHost(nameIndex, length, file, constantPool);
            }
            case 26: {
                return new NestMembers(nameIndex, length, file, constantPool);
            }
            default: {
                throw new IllegalStateException("Unrecognized attribute type tag parsed: " + tag);
            }
        }
    }
    
    public static Attribute readAttribute(final DataInputStream file, final ConstantPool constantPool) throws IOException {
        return readAttribute((DataInput)file, constantPool);
    }
    
    public static void removeAttributeReader(final String name) {
        Attribute.READERS.remove(name);
    }
    
    protected Attribute(final byte tag, final int nameIndex, final int length, final ConstantPool constantPool) {
        this.tag = tag;
        this.name_index = nameIndex;
        this.length = length;
        this.constant_pool = constantPool;
    }
    
    @Override
    public abstract void accept(final Visitor p0);
    
    public Object clone() {
        Attribute attr = null;
        try {
            attr = (Attribute)super.clone();
        }
        catch (final CloneNotSupportedException e) {
            throw new Error("Clone Not Supported");
        }
        return attr;
    }
    
    public abstract Attribute copy(final ConstantPool p0);
    
    public void dump(final DataOutputStream file) throws IOException {
        file.writeShort(this.name_index);
        file.writeInt(this.length);
    }
    
    public final ConstantPool getConstantPool() {
        return this.constant_pool;
    }
    
    public final int getLength() {
        return this.length;
    }
    
    public String getName() {
        return this.constant_pool.getConstantUtf8(this.name_index).getBytes();
    }
    
    public final int getNameIndex() {
        return this.name_index;
    }
    
    public final byte getTag() {
        return this.tag;
    }
    
    public final void setConstantPool(final ConstantPool constantPool) {
        this.constant_pool = constantPool;
    }
    
    public final void setLength(final int length) {
        this.length = length;
    }
    
    public final void setNameIndex(final int nameIndex) {
        this.name_index = nameIndex;
    }
    
    @Override
    public String toString() {
        return Const.getAttributeName(this.tag);
    }
    
    static {
        debug = Boolean.getBoolean(Attribute.class.getCanonicalName() + ".debug");
        READERS = new HashMap<String, Object>();
        EMPTY_ARRAY = new Attribute[0];
    }
}
